package pipeline

import (
	"testing"

	"github.com/stretchr/testify/assert"
	yaml "gopkg.in/yaml.v2"
)

func loadPath(s string) Path {
	var p Path
	if err := yaml.Unmarshal([]byte(s), &p); err != nil {
		panic(err)
	}
	return p
}
func loadPathList(s string) PathList {
	var p PathList
	if err := yaml.Unmarshal([]byte(s), &p); err != nil {
		panic(err)
	}
	return p
}

func TestContainsCycle(t *testing.T) {
	t.Run("EmptyPathList", func(t *testing.T) {
		p := loadPath(`- src -> dest`)
		l := loadPathList(`[]`)
		assert.False(t, l.ContainsCycle(p))
	})
	t.Run("NoMatch", func(t *testing.T) {
		p := loadPath(`[ src -> dest, dest -> src ]`)
		l := loadPathList(`- [ src -> relay, relay -> dest ]`)
		assert.False(t, l.ContainsCycle(p))
	})
	t.Run("Match", func(t *testing.T) {
		p := loadPath(`[ src -> relay, relay -> dest ]`)
		l := loadPathList(`- [ src -> relay, relay1 -> relay, relay -> dest ]
- [ src -> relay, src -> relay ]
- [ src -> relay, relay -> dest ]`)
		assert.True(t, l.ContainsCycle(p))
	})
}

func TestValidator(t *testing.T) {
	var ValidatorTests = []struct {
		name      string
		graph     string
		errorMsgs []string
	}{
		{"emptygraph", "", []string{
			"Graph must have one emitter (found none)",
			"Graph must have at least one consumer (found none)",
		}},

		{"disconnected output & no consumer", `
nodes:
  - id: src
    type: emitter
    sourcematch:
      eventtypes: ["default "]
    outputs: [default]
`,
			[]string{
				"Emitter 'src': invalid event type 'default '",
				"Graph must have at least one consumer (found none)",
				"Output src.default is not connected by any edge",
			},
		},

		{"disconnected input & no emitter", `
nodes:
  - id: out
    type: consumer
    inputs: [default]
    roles: [role]
    rolebroadcast: true
`,
			[]string{
				"Graph must have one emitter (found none)",
				"Input out.default is not connected by any edge",
			},
		},

		{"invalid edges", `
nodes:
  - id: in
    type: emitter
    sourcematch:
      eventtypes: [default]
    outputs: [default]
  - id: out
    type: consumer
    inputs: [default]
    roles: [role]
    rolebroadcast: true
edges:
  - in.default -> out.default
  - in.output2 -> out.default
  - out.output1 -> in.default
`,
			[]string{
				"Edge in.output2->out.default origin does not exist",
				"Edge out.output1->in.default destination does not exist",
				"Edge out.output1->in.default origin does not exist",
			},
		},

		{"loop", `
nodes:
- id: in
  type: emitter
  sourcematch:
    eventtypes: [default]
  outputs: [default]
- id: w
  type: worker
  inputs: [in]
  outputs: [out]
- id: out
  type: consumer
  inputs: [default]
  roles: [role]
  rolebroadcast: true
edges:
- in.default -> w.in
- w.out -> w.in
- w.out -> out.default
`,
			[]string{
				"Cycle detected: [w.out->w.in]",
			},
		},

		{"disconnected graph", `
nodes:
- id: in1
  type: emitter
  sourcematch:
    eventtypes: [default]
  outputs: [default]
- id: out1
  type: consumer
  inputs: [default]
  roles: [role]
  rolebroadcast: true
- id: in2
  type: emitter
  sourcematch:
    eventtypes: [default]
  outputs: [default]
- id: out2
  type: consumer
  inputs: [default]
  roles: [role]
  rolebroadcast: true
edges:
- in1->out1
- in2->out2
`,
			[]string{
				"Found extra emitter node: in2",
				"The graph is not connected",
			},
		},
		{"missing sourcematch", `
nodes:
- id: in
  type: emitter
  outputs: [default]
- id: out
  type: consumer
  inputs: [default]
  roles: [role]
  rolebroadcast: true
edges:
- in->out
`,
			[]string{
				"Emitter 'in' has no sourcematch",
			},
		},
		{"bad emitter/consumer input/output", `
nodes:
- id: in
  type: emitter
  sourcematch:
    eventtypes: [default]
  inputs: [default]
- id: out
  type: consumer
  outputs: [default]
  roles: [role]
  rolebroadcast: true
edges:
- in->out
`,
			[]string{
				"Emitter 'in' has no output. It must have one",
				"Emitter 'in' has some input. It must not have any",
				"Consumer 'out' has no input. It must have one",
				"Consumer 'out' has some output. It must not have any",
				"Edge in.default->out.default destination does not exist",
				"Edge in.default->out.default origin does not exist",
				"Input in.default is not connected by any edge",
				"Output out.default is not connected by any edge",
			},
		},
		{"emitter/consumer multiple input/output", `
nodes:
- id: in
  type: emitter
  sourcematch:
    eventtypes: [default]
  outputs: [default, o2]
- id: out
  type: consumer
  inputs: [default, i2]
  roles: [role]
  rolebroadcast: true
edges:
- in->out
- in.o2->out.i2
`,
			[]string{
				"Emitter 'in' has multiple output. It must have only one",
				"Consumer 'out' has multiple input. It must have only one",
			},
		},
		{"roles/broadcast", `
nodes:
- id: in1
  type: emitter
  sourcematch:
    eventtypes: [default]
  roles: [role1, role2]
  outputs: [default]
- id: out1
  type: consumer
  roles: [role1]
  rolebroadcast: false
  inputs: [default]
edges:
- in1->out1
`,
			[]string{
				"Node 'in1' has more than 1 role",
				"Node 'in1' should have rolebroadcast enabled",
				"Node 'out1' should have rolebroadcast enabled",
			},
		},
		{"consumer w multiple roles", `
nodes:
- id: in1
  type: emitter
  sourcematch:
    eventtypes: [default]
  roles: [role1, role2]
  rolebroadcast: true
  outputs: [default]
- id: out1
  type: consumer
  roles: [role1]
  rolebroadcast: true
  actors: [actor1]
  inputs: [default]
edges:
- in1->out1
`,
			[]string{
				"Node 'in1' has more than 1 role",
			}},
		{"consumer w actorids", `
nodes:
- id: in1
  type: emitter
  sourcematch:
    eventtypes: [default]
  roles: [role1]
  rolebroadcast: true
  outputs: [default]
- id: out1
  type: consumer
  actorids: ["db40bea4-61ae-4be7-9fd3-94c937d9ac5e"]
  actors: [actor1]
  inputs: [default]
edges:
- in1->out1
`,
			[]string{
				"Consumer 'out1' cannot have 'actorids' defined",
			}},
		{"invalid source/sink types", `
nodes:
- id: in1
  type: worker
  outputs: [default]
- id: out1
  type: worker
  inputs: [i1, i2]
  roles: [role]
  rolebroadcast: true
- id: in2
  type: emitter
  sourcematch:
    eventtypes: [default]
  outputs: [default]
- id: out2
  type: consumer
  inputs: [default]
  roles: [role]
  rolebroadcast: true
edges:
- in1->out1.i1
- in2->out2
- in2->out1.i2
`,
			[]string{
				"Source node 'in1' is not an emitter",
				"Sink node 'out1' is not a consumer",
			},
		},
		{"invalid id/duplicate id", `
nodes:
- id: in 1
  type: emitter
  sourcematch:
    eventtypes: [default]
  outputs: [default]
- id: worker#1
  type: worker
  inputs: [default]
  outputs: [default]
- id: w2
  type: worker
  inputs: [default]
  outputs: [default]
- id: w2
  type: worker
  inputs: [default]
  outputs: [default]
- id:
  type: consumer
  inputs: [default]
edges:
- in 1->worker#1
- worker#1->w2
- w2->.default
  `,
			[]string{
				"Node ID 'in 1' is not a valid ID. It should match the regexp '^[a-zA-Z0-9_\\-]+$'",
				"Node ID 'worker#1' is not a valid ID. It should match the regexp '^[a-zA-Z0-9_\\-]+$'",
				"Two or more nodes have the same ID 'w2'",
				"Node ID '' is not a valid ID. It should match the regexp '^[a-zA-Z0-9_\\-]+$'",
				"Consumer '' has no 'roles' nor 'actors'",
				"Cycle detected: [.->. .->. w2.default->.default]",
				"The graph is not connected",
			},
		},
	}
	for _, test := range ValidatorTests {
		t.Run(test.name, func(t *testing.T) {
			var pl Pipeline
			if err := yaml.Unmarshal([]byte(test.graph), &pl.Graph); err != nil {
				t.Fatal(err)
			}
			validator := NewValidator(pl.Graph)
			errors := validator.Validate()
			assert.Equal(t, test.errorMsgs, errors.Messages())
		})
	}
}
