// Copyright 2016 Orus.io. All rights reserved.

package registration

import (
	"context"
	"encoding/pem"
	"fmt"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/go-xbus/tools"
	"bitbucket.org/orus-io/xbus/services"
	"bitbucket.org/orus-io/xbus/storage"
)

// QueueName used for queue subscriptions
const QueueName = "xbus.services.registration"

// Options of the Service
type Options struct {
	services.Options

	AccountStorage storage.AccountStorage
	ActorStorage   storage.ActorStorage
	ServerCA       []byte
}

// NewService instanciate a Service
func NewService(o Options) *Service {
	return &Service{
		o:           o,
		BaseService: services.NewBaseService("registration", o.Options),
	}
}

// Service implements the registration api
type Service struct {
	services.BaseService
	o Options
}

// Startup subscribes the service to the registration apis
func (s *Service) Startup() error {
	regHandler := api.NewSvcRegistrationHandler(
		context.Background(), s.Conn, registrationServer{s},
	)
	if err := s.SubscribeAll(
		s.HandlerQueueSubscribe(regHandler, QueueName),
	); err != nil {
		s.Shutdown()
		return err
	}
	return nil
}

type registrationServer struct {
	*Service
}

func (s registrationServer) Register(
	ctx context.Context, req api.RegistrationRequest) (api.Registration, error) {
	return s.handleRegistration(api.UUID{}, req)
}

func (s registrationServer) Update(
	ctx context.Context, accountID string, req api.RegistrationRequest) (api.Registration, error) {
	accountUUID, err := api.UUIDFromString(accountID)
	if err != nil {
		return api.Registration{}, fmt.Errorf("Invalid accountID: %s", err)
	}
	return s.handleRegistration(accountUUID, req)
}

func parseSubject(subject string) (anon bool, accountID api.UUID) {
	if subject == api.RegistrationSubject {
		anon = true
		return
	}
	if len(subject) != len(api.RegistrationSubject)+37 {
		panic("Invalid registration subject: " + subject)
	}
	strID := subject[len(api.RegistrationSubject):]
	return false, api.MustUUIDFromString(strID)
}

func (s *Service) checkActors(accountID api.UUID, actors []api.Actor) error {
	// For each actor, if not-yet created (ie if this is an original req),
	// create the actor in the pending state
	for _, actor := range actors {
		if a, exists := s.o.ActorStorage.GetByName(actor.Name); exists {
			if accountID.IsZero() || a.AccountID != accountID {
				return fmt.Errorf(
					"%s: Actor %s is attached to a different account (%s)",
					accountID, a.Name, a.AccountID)
			}
		}
	}
	return nil
}

func (s *Service) handleRegistration(
	accountID api.UUID, req api.RegistrationRequest,
) (api.Registration, error) {
	var account api.Account

	if accountID.IsZero() {
		var (
			ok  bool
			err error
		)
		if !tools.ValidateAccountName(req.Name) {
			return api.Registration{}, fmt.Errorf("Invalid account name: '%s'", req.Name)
		}
		account, ok = s.o.AccountStorage.GetByName(req.Name)
		if !ok {
			if err := s.checkActors(accountID, req.ActorList); err != nil {
				return api.Registration{}, err
			}
			// If non-existing account, create it in the adequate state and its csr attached
			account = api.Account{
				Name:   req.Name,
				Type:   req.Type,
				Status: api.Account_PENDING,
			}
			if req.CSR != "" {
				account.CSR = req.CSR
				account.CSROrigin = api.Account_ANONYMOUS
			}
			account, err = s.o.AccountStorage.Create(account)
			if err != nil {
				return api.Registration{}, api.ServerProtoError(err)
			}
		} else {
			s.Log.Noticef("%#v  -  %#v", req, account)
			if err := s.checkActors(account.ID, req.ActorList); err != nil {
				return api.Registration{}, err
			}
			if account.CSR != "" && req.CSR != account.CSR {
				return api.Registration{}, fmt.Errorf("CSR validation already pending. Accept or reject first")
			}
			if account.CSR == "" && account.Status == api.Account_PENDING {
				account.CSR = req.CSR
				account.CSROrigin = api.Account_ANONYMOUS
				s.Log.Noticef("Setting CSR on the account")
				var err error
				account, err = s.o.AccountStorage.Update(account)
				if err != nil {
					return api.Registration{}, err
				}
			}
		}
	} else {
		var exists bool
		account, exists = s.o.AccountStorage.Get(accountID)
		if !exists {
			return api.Registration{}, api.ClientProtoError(
				fmt.Sprintf("No such account %s", account),
			)
		}
		if err := s.checkActors(account.ID, req.ActorList); err != nil {
			return api.Registration{}, err
		}
		if req.CSR != "" {
			if account.Status != api.Account_PENDING {
				return api.Registration{}, fmt.Errorf("To renew an account cert, you must first run 'xbusctl account reset-cert <account>'")

			}
			if account.CSR != "" && account.CSR != req.CSR {
				return api.Registration{}, fmt.Errorf("CSR validation already pending. Accept or reject first")
			}
			account.CSR = req.CSR
			account.CSROrigin = api.Account_AUTHACCOUNT
			var err error
			account, err = s.o.AccountStorage.Update(account)
			if err != nil {
				return api.Registration{}, fmt.Errorf("Failed to update account: %s", err)
			}
		}
	}

	// For each actor, if not-yet created (ie if this is an original req),
	// create the actor in the pending state
	for _, actor := range req.ActorList {
		if _, exists := s.o.ActorStorage.GetByName(actor.Name); !exists {
			sActor := api.Actor{
				Name:      actor.Name,
				Kind:      actor.Kind,
				AccountID: account.ID,
				Status:    api.Actor_PENDING,
				Roles:     actor.Roles,
			}
			var err error
			sActor, err = s.o.ActorStorage.Create(sActor)
			if err != nil {
				return api.Registration{}, api.ClientProtoError(
					"Error creating Actor '%s': %s",
					actor.Name, err,
				)
			}
		}
	}

	// Check the current state of the account and actors and put it all
	// in the reply, along with the signed certificate if applicable
	reg := api.Registration{
		ID: account.ID,
	}
	switch account.Status {
	case api.Account_PENDING:
		reg.RegistrationStatus = api.Registration_PENDING
	case api.Account_ENABLED:
		reg.RegistrationStatus = api.Registration_ACCEPTED
	case api.Account_DISABLED:
		if account.Cert == "" {
			reg.RegistrationStatus = api.Registration_REJECTED
		} else {
			reg.RegistrationStatus = api.Registration_ACCEPTED
		}
	}
	if reg.RegistrationStatus == api.Registration_ACCEPTED {
		reg.SignedCertificate = account.Cert
		if len(s.o.ServerCA) != 0 {
			reg.ServerCA = string(pem.EncodeToMemory(&pem.Block{
				Type:  "CERTIFICATE",
				Bytes: s.o.ServerCA,
			}))
		}
	}

	actorList := s.o.ActorStorage.ListAccount(account.ID)

	for _, actor := range actorList {
		reg.ActorList = append(reg.ActorList, api.Actor{
			ID:     actor.ID,
			Name:   actor.Name,
			Kind:   actor.Kind,
			Status: actor.Status,
			Roles:  actor.Roles,
		})
	}
	return reg, nil
}
