package tools

import (
	"context"
	"sync"

	nats "github.com/nats-io/go-nats"

	xbus "bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
)

// ActorState is emitted by the ActorMap when a actor status changes
type ActorState struct {
	ID     api.UUID
	Name   string
	Kind   api.Actor_Kind
	Online bool
}

// ActorMap holds a map of actors->status and keep it in sync by subscribing
// to actor events
type ActorMap struct {
	ctx    context.Context
	cancel func()

	m  sync.Mutex
	wg sync.WaitGroup

	log xbus.Logger

	setOnlineChan chan api.CoreEvent
	eventChan     <-chan api.CoreEvent
	eventSub      *nats.Subscription

	actors map[api.UUID]ActorState

	Events chan ActorState
}

// NewActorMap initialize the map and starts listening
func NewActorMap(ctx context.Context, client *api.Client, log xbus.Logger) (*ActorMap, error) {
	am := ActorMap{
		actors: make(map[api.UUID]ActorState),
		log:    log,

		// We may need a buffer here to make
		// sure the incoming events are handled asap
		Events: make(chan ActorState),
	}
	am.ctx, am.cancel = context.WithCancel(ctx)

	var err error
	if am.eventChan, am.eventSub, err = client.CoreEvents.ActorSubscribeChan(); err != nil {
		return nil, err
	}

	am.setOnlineChan = make(chan api.CoreEvent)

	var eventQueue []api.CoreEvent
	stopQueing := make(chan struct{})

	go func() {
		for {
			select {
			case <-am.ctx.Done():
				return
			case <-stopQueing:
				return
			case event := <-am.eventChan:
				eventQueue = append(eventQueue, event)
			}
		}
	}()

	// fetch the actors from the sentinel
	err = client.Sentinel.ListActors(ctx,
		func(ctx context.Context, reply api.ClientGetActorsReply) {
			am.m.Lock()
			defer am.m.Unlock()
			for _, actor := range reply.Actors {
				am.actors[actor.ID] = ActorState{
					ID:     actor.ID,
					Name:   actor.Name,
					Kind:   actor.Kind,
					Online: actor.Online,
				}
			}
		},
	)
	close(stopQueing)

	if err != nil {
		return nil, err
	}

	// apply pending actor events
	for _, event := range eventQueue {
		am.applyEvent(event)
	}

	// start syncing
	am.wg.Add(1)
	go am.run()

	return &am, nil
}

func (am *ActorMap) run() {
	defer am.wg.Done()
	defer func() {
		if err := am.eventSub.Unsubscribe(); err != nil {
			am.log.Errorf("Error unsubscribing from actor events: %s", err)
		}
		am.eventChan = nil
	}()

	for {
		select {
		case <-am.ctx.Done():
			return
		case event := <-am.setOnlineChan:
			ev := am.applyEvent(event)
			if ev != nil {
				am.Events <- *ev
			}
		case event := <-am.eventChan:
			ev := am.applyEvent(event)
			if ev != nil {
				am.Events <- *ev
			}
		}
	}
}

func (am *ActorMap) applyEvent(event api.CoreEvent) *ActorState {
	switch event.Name {
	case "up":
		return am.setOnline(event.Actor, true)
	case "down":
		return am.setOnline(event.Actor, false)
	default:
		return nil
	}
}

func (am *ActorMap) setOnline(actor *api.Actor, online bool) *ActorState {
	am.m.Lock()
	defer am.m.Unlock()

	actorState, ok := am.actors[actor.ID]
	if !ok {
		actorState = ActorState{
			ID:   actor.ID,
			Name: actor.Name,
			Kind: actor.Kind,
		}
	} else if actorState.Online == online {
		return nil
	}
	actorState.Online = online

	am.actors[actor.ID] = actorState
	return &actorState
}

// SetOnline force the status of an actor
func (am *ActorMap) SetOnline(id api.UUID, online bool) {
	ename := "down"
	if online {
		ename = "up"
	}
	am.setOnlineChan <- api.CoreEvent{
		Name:  ename,
		Actor: &api.Actor{ID: id},
	}
}

// IsOnline returns the status of an actor
func (am *ActorMap) IsOnline(id api.UUID) (bool, bool) {
	am.m.Lock()
	defer am.m.Unlock()

	actor, ok := am.actors[id]
	return actor.Online, ok
}

// ListOnline returns a list of online actors
func (am *ActorMap) ListOnline() []api.UUID {
	am.m.Lock()
	defer am.m.Unlock()

	var list []api.UUID
	for id, actor := range am.actors {
		if actor.Online {
			list = append(list, id)
		}
	}
	return list
}

// List returns the list of actors
func (am *ActorMap) List() (list []ActorState) {
	am.m.Lock()
	defer am.m.Unlock()

	for _, actor := range am.actors {
		list = append(list, actor)
	}
	return list
}

// Close stops the sync
func (am *ActorMap) Close() {
	am.cancel()
	am.cancel = nil
	am.wg.Wait()
}
