// Copyright 2017 Orus.io. All rights reserved.

package xbus

import (
	"context"
	"fmt"
	"runtime/debug"
	"time"

	nats "github.com/nats-io/go-nats"
	nrpc "github.com/nats-rpc/gogo-nrpc"

	api "bitbucket.org/orus-io/go-xbus/api"
)

// EnvelopeSenderDefaultDelay is the default MaxDelay of the envelope senders
// created by Actor.Output()
var EnvelopeSenderDefaultDelay = 10 * time.Second

func catchErrors(log Logger) {
	if r := recover(); r != nil {
		log.Errorf("Caught error: %s\n%s", r, debug.Stack())
	}
}

// ActorProcessHandler is a function that implements a Actor.Process
type ActorProcessHandler func(*ActorProcessingContext) error

// newActor initialize an actor
func newActor(client *Client, options ActorOptions) *Actor {
	actor := Actor{
		ID:      options.ID,
		Name:    options.Name,
		Kind:    options.Kind,
		Options: options,
		Client:  client,
	}
	actor.Log = &ActorLogger{
		NewPrefixedLogger("[actor/"+options.Name+"] ", client.Log),
		&actor}
	return &actor
}

// Actor provides the common tools for all actors
type Actor struct {
	ID      api.UUID
	Name    string
	Kind    api.Actor_Kind
	Options ActorOptions
	Log     *ActorLogger
	Client  *Client
}

// GetStringSetting returns a string typed setting entry
func (a *Actor) GetStringSetting(name string) string {
	value, ok := a.Options.Service.Settings[name]
	if !ok {
		return ""
	}
	sValue, ok := value.(string)
	if !ok {
		panic(fmt.Sprintf("Invalid configuration value: %s", value))
	}
	return sValue
}

// MustGetStringSetting returns a string setting or panic if undefined
func (a *Actor) MustGetStringSetting(name string) string {
	value := a.GetStringSetting(name)
	if value == "" {
		a.Client.Log.Fatalf("Missing setting: %s", name)
	}
	return value
}

// GetIntSettingD returns a int setting, or the defaultValue if not specified
func (a *Actor) GetIntSettingD(name string, defaultValue int) int {
	value, ok := a.Options.Service.Settings[name]
	if !ok {
		return defaultValue
	}
	iValue, ok := value.(int)
	if !ok {
		panic(fmt.Sprintf("Invalid configuration value '%s'. Expected an integer", value))
	}
	return iValue
}

// GetBoolSettingD returns a boolean setting, or the defaultValue if not specified
func (a *Actor) GetBoolSettingD(name string, defaultValue bool) bool {
	value, ok := a.Options.Service.Settings[name]
	if !ok {
		return defaultValue
	}
	bValue, ok := value.(bool)
	if !ok {
		panic(fmt.Errorf("Invalid configuration value '%s'. Expected a boolean", value))
	}
	return bValue
}

// GetStringSliceSetting returns a []string setting, or nil if undefined
func (a *Actor) GetStringSliceSetting(name string) []string {
	value, ok := a.Options.Service.Settings[name]
	if !ok {
		return nil
	}
	sValue, ok := value.([]string)
	if !ok {
		siValue, ok := value.([]interface{})
		if !ok {
			panic(fmt.Errorf("Invalid configuration value '%#v'. Expected an array of strings", value))
		}
		for _, i := range siValue {
			s, ok := i.(string)
			if !ok {
				panic(fmt.Errorf("Invalid configuration entry '%#v'. Expected a string", i))
			}
			sValue = append(sValue, s)
		}
	}
	return sValue
}

// GetClient returns the underlying Client instance
func (a *Actor) GetClient() *Client {
	return a.Client
}

// ProcessingEnd calls the ActorProcessingEnd xbus api
func (a *Actor) ProcessingEnd(
	context *api.ProcessingContext,
	status api.ActorProcessingState_Status,
	messages []api.LogMessage,
) error {
	_, err := a.GetClient().tryProcessingEnd(
		a.ID, *context, status, messages, true)

	return err
}

// ProcessingSuccess signal the success of a processing
// for actors with no outputs of it not all outputs were
// hit or closed
func (a *Actor) ProcessingSuccess(context *api.ProcessingContext) error {
	return a.ProcessingEnd(context, api.ActorProcessingState_SUCCESS, nil)
}

// ProcessingError signals a processing error.
func (a *Actor) ProcessingError(context *api.ProcessingContext, format string, fmtArgs ...interface{}) error {
	return a.ProcessingEnd(context, api.ActorProcessingState_ERROR, []api.LogMessage{
		api.LogMessage{
			Time:  time.Now(),
			Level: api.LogLevel_ERROR,
			Text:  fmt.Sprintf(format, fmtArgs...),
		}})
}

// SendEnvelope send a potentially incomplete envelope to the broker
func (a *Actor) SendEnvelope(ctx *api.ProcessingContext, output string, env api.Envelope) (status api.EnvelopeAck_ReceptionStatus, err error) {
	if a.Kind == api.Actor_CONSUMER {
		return api.EnvelopeAck_ERROR, fmt.Errorf("A 'Consumer' actor cannot emit to outputs")
	}
	return a.Client.SendEnvelope(a.ID, ctx, output, env)
}

// SendItems send a group of items to the broker
// The envelope and event will be constructed automatically
func (a *Actor) SendItems(ctx *api.ProcessingContext, output, eventType string, items []api.Item) (status api.EnvelopeAck_ReceptionStatus, err error) {
	return a.SendEnvelope(ctx, output, ForgeEnvelopeFromItems(eventType, items))
}

// ForgeEnvelopeFromEvent ...
func ForgeEnvelopeFromEvent(event api.Event) api.Envelope {
	envID, err := api.UUIDV4()
	if err != nil {
		panic(err)
	}
	return api.Envelope{
		ID:       envID,
		EventIDs: []api.UUID{event.ID},
		Events:   []api.Event{event},
	}
}

// ForgeEnvelopeFromItems ...
func ForgeEnvelopeFromItems(eventType string, items []api.Item) api.Envelope {
	evID, err := api.UUIDV4()
	if err != nil {
		panic(err)
	}
	checksum := api.Checksum(0).
		Update(evID[:16]).
		Update([]byte(eventType))

	for _, item := range items {
		checksum = checksum.Update(item)
	}

	event := api.Event{
		ID:        evID,
		Type:      eventType,
		Index:     1,
		Items:     items,
		ItemCount: uint64(len(items)),
		Checksum:  checksum,
	}
	return ForgeEnvelopeFromEvent(event)
}

type actorAgentServer struct {
	actor  *Actor
	handle ActorProcessHandler
}

func (s actorAgentServer) Process(
	context context.Context, req api.ActorProcessRequest,
	send func(api.ActorProcessingState),
) error {
	apc := ActorProcessingContext{
		Request:  &req,
		Context:  context,
		Actor:    s.actor,
		Log:      s.actor.Log,
		LocalLog: s.actor.Client.Log,
		currentState: api.ActorProcessingState{
			Context: *req.Context,
			Status:  api.ActorProcessingState_RUNNING,
		},
		send:                 send,
		retrieveEnvelopeFrom: s.actor.Client.RetrieveEnvelopeFrom,
		openOutput:           s.actor.OpenOutput,
		closeOutput:          s.actor.CloseOutput,
		sendEnvelope:         s.actor.SendEnvelope,
		processingEnd:        s.actor.ProcessingEnd,
	}
	if err := s.handle(&apc); err != nil {
		if apc.done {
			// The handler returned an error _and_ called Done() or Error().
			s.actor.Log.Errorf(
				"the handler returned an error after calling Error() or Done(), "+
					"which is a forbidden behavior. Please check the handler "+
					"implementation. The state that was sent is: %v, and the "+
					"returned error is '%s'", apc.currentState, err,
			)
		} else if apc.detached {
			// The handler returned an error _and_ called Detached()
			s.actor.Log.Errorf(
				"the handler returned an error after calling Detached(), "+
					"which is a forbidden behavior. Please check the handler "+
					"implementation. The returned error is '%s'", err,
			)
		} else if _, ok := err.(*nrpc.Error); ok {
			return err
		} else {
			apc.Error(err)
		}
	} else if !apc.detached && !apc.done {
		apc.Done()
	}
	return nil
}

// Subscription is the actor subscription for handling incoming process requests
type Subscription struct {
	processSub *nats.Subscription
	surveySub  *nats.Subscription
	actor      *Actor
	workers    *nrpc.WorkerPool
}

func (s Subscription) handleSurvey(request api.SurveyRequest) {
	defer catchErrors(s.actor.Client.Log)

	if request.OnlineStatus {
		if err := s.actor.Client.API.Sentinel.ActorReady(s.actor.ID.String()); err != nil {
			s.actor.Client.Log.Errorf("Error calling 'ActorReady': %s", err)
		}
	}
}

// Unsubscribe the subscription
func (s Subscription) Unsubscribe() error {
	if err := s.actor.Client.API.Sentinel.ActorLeaving(s.actor.ID.String()); err != nil {
		s.actor.Client.Log.Errorf("Error calling 'ActorLeaving': %s", err)
	}
	s.actor.Client.Log.Noticef("ActorLeaving called")
	err, err2 := s.processSub.Unsubscribe(), s.surveySub.Unsubscribe()
	if err == nil {
		err = err2
	}
	s.workers.Close(time.Second)
	return err
}

// Subscribe registers a ActorAgent implementation and starts listening
func (a *Actor) Subscribe(
	handle ActorProcessHandler,
) (*Subscription, error) {
	conn := a.Client.GetConn()

	workers := nrpc.NewWorkerPool(
		context.Background(), 1, 0, time.Second)

	handler := api.NewActorAgentConcurrentHandler(
		workers,
		conn,
		actorAgentServer{
			actor:  a,
			handle: handle,
		},
	)
	// TODO keep track of active calls so the actor cannot be shutdown
	// until all current calls are canceled or ended (we should use context
	// for the canceling)
	sub, err := conn.QueueSubscribe(
		api.XbusAPIPrefix+"actoragent."+a.ID.String()+".>",
		a.ID.String(),
		handler.Handler,
	)
	if err != nil {
		workers.Close(0)
		return nil, err
	}
	subscription := &Subscription{
		processSub: sub,
		actor:      a,
		workers:    workers,
	}

	if subscription.surveySub, err = a.Client.API.Sentinel.SurveySubscribe(
		subscription.handleSurvey,
	); err != nil {
		sub.Unsubscribe()
		return nil, err
	}

	if err := a.Client.API.Sentinel.ActorReady(a.ID.String()); err != nil {
		sub.Unsubscribe()
		subscription.surveySub.Unsubscribe()
		return nil, err
	}

	return subscription, nil
}

// CloseOutput closes an output
func (a *Actor) CloseOutput(
	context *api.ProcessingContext, output string,
) (api.EnvelopeAck, error) {
	return a.Client.sendEnvelope(a.ID, context, output, nil)
}

// OpenOutput returns a EnvelopeSender
func (a *Actor) OpenOutput(
	context *api.ProcessingContext, output string, eventTypes ...string,
) *EnvelopeSender {
	var es *EnvelopeSender
	es = NewEnvelopeSender(
		eventTypes, api.EnvelopeMaxSize,
		func(env *api.Envelope) (api.EnvelopeAck_ReceptionStatus, error) {
			ack, err := a.Client.sendEnvelope(a.ID, context, output, env)
			return ack.GetStatus(), err
		})
	es.SetMaxDelay(EnvelopeSenderDefaultDelay)
	return es
}
