package api

import (
	"bytes"
	"encoding/binary"
	"errors"

	muuid "github.com/m4rw3r/uuid"
)

// UUID is thin wrapper of m4rw3r/uuid.UUID that implements the needed
// interfaces for being used as gogo/protobuf
type UUID muuid.UUID

// IsZero returns true if the uuid has only zeros
func (uuid UUID) IsZero() bool {
	return muuid.UUID(uuid).IsZero()
}

// SetZero reset the uuid to zeros
func (uuid *UUID) SetZero() {
	(*muuid.UUID)(uuid).SetZero()
}

// String returns a string (hex) representation of the uuid
func (uuid UUID) String() string {
	return muuid.UUID(uuid).String()
}

// SetString sets the uuid value from a string
func (uuid *UUID) SetString(str string) error {
	return (*muuid.UUID)(uuid).SetString(str)
}

// ReadBytes reads the supplied byte array of hexadecimal characters representing
// a UUID into the instance.
func (uuid *UUID) ReadBytes(str []byte) error {
	return (*muuid.UUID)(uuid).ReadBytes(str)
}

// Marshal returns the bytes representation for protobuf encoding
func (uuid UUID) Marshal() ([]byte, error) {
	if len(uuid) == 0 {
		return nil, nil
	}
	return uuid[:16], nil
}

// MarshalTo write the bytes representation for protobuf encoding
// to a buffer
func (uuid UUID) MarshalTo(data []byte) (n int, err error) {
	if len(uuid) == 0 {
		return 0, nil
	}
	copy(data, uuid[:16])
	return 16, nil
}

// Unmarshal loads a value from a protobuf encoding
func (uuid *UUID) Unmarshal(data []byte) error {
	if len(data) == 0 {
		uuid = nil
		return nil
	}
	if len(data) != 16 {
		return errors.New("Invalid UUID")
	}
	copy((*uuid)[:16], data)
	return nil
}

// Size returns the size needed for a protobuf encoding
func (uuid *UUID) Size() int {
	if uuid == nil {
		return 0
	}
	if len(*uuid) == 0 {
		return 0
	}
	return 16
}

// MarshalText returns the string-representation of the UUID as a byte-array.
func (uuid UUID) MarshalText() ([]byte, error) {
	return muuid.UUID(uuid).MarshalText()
}

// UnmarshalText reads an UUID from a string into the UUID instance.
// If this fails the state of the UUID is undetermined.
func (uuid *UUID) UnmarshalText(data []byte) error {
	return (*muuid.UUID)(uuid).UnmarshalText(data)
}

// MarshalJSON returns the string-representation of the UUID as a JSON-string.
func (uuid UUID) MarshalJSON() ([]byte, error) {
	return muuid.UUID(uuid).MarshalJSON()
}

// UnmarshalJSON reads an UUID from a JSON-string into the UUID instance.
// If this fails the state of the UUID is undetermined.
func (uuid *UUID) UnmarshalJSON(data []byte) error {
	return (*muuid.UUID)(uuid).UnmarshalJSON(data)
}

// Equal returns true if the two values are equal. Needed by gogo/protobuf
func (uuid UUID) Equal(other UUID) bool {
	return uuid == other
}

// Compare compares two value
func (uuid UUID) Compare(other UUID) int {
	return bytes.Compare(uuid[:16], other[:16])
}

type randyUUID interface {
	Uint32() uint32
}

// NewPopulatedUUID implements V4 with a randy source
func NewPopulatedUUID(r randyUUID) *UUID {
	u := UUID{}

	binary.LittleEndian.PutUint32(u[0:4], r.Uint32())
	binary.LittleEndian.PutUint32(u[4:8], r.Uint32())
	binary.LittleEndian.PutUint32(u[8:12], r.Uint32())
	binary.LittleEndian.PutUint32(u[12:16], r.Uint32())

	u[8] = (u[8] | 0x80) & 0xBF
	u[6] = (u[6] | 0x40) & 0x4F

	return &u
}

// MaybeUUIDFromString reads a UUID into a new UUID instance,
// setting the instance to zero if it fails.
func MaybeUUIDFromString(s string) UUID {
	return UUID(muuid.MaybeFromString(s))
}

// MustUUIDFromString parses and returns a UUID, and panics if it fails to do so
func MustUUIDFromString(s string) UUID {
	return UUID(muuid.MustFromString(s))
}

// UUIDFromString parses and returns a UUID
func UUIDFromString(s string) (UUID, error) {
	u, err := muuid.FromString(s)
	return UUID(u), err
}

// UUIDV4 creates a UUID from rand
func UUIDV4() (UUID, error) {
	u, err := muuid.V4()
	return UUID(u), err
}
